import numpy as np
import numba as nb
from numba import jit, objmode, types
from scipy.stats import rankdata
from scipy.spatial.distance import cdist
from sklearn.preprocessing import MinMaxScaler
import matplotlib.pyplot as plt
from .MOF import _Var_Massratio
from .MOF import _point_in_radius

# Calculate windowing mass ratio
@jit(nopython=True)
def _Window_Massratio(Data,Window,Overlap_ratio):
  n =Data.shape[0]
  scores = np.zeros(n)

  assert(Window > 0 )
  assert(0.0 <= Overlap_ratio <= 0.5)

  overlap_size = int(Overlap_ratio * Window)
  mid_overlap_point = int(Window - overlap_size/2)
  score_count = 0

  # Windowing data and send to MOF
  for start_point in range(0, n , Window - overlap_size):
    stop_point = min(n, start_point + Window)
    w = stop_point - start_point
    current_data = Data[start_point : stop_point]
    # Last data window size (w) <= Window
    current_scores = _Var_Massratio(current_data,w)

    # Assign score to data points
    stop_score_count = min(n, start_point + mid_overlap_point)
    for i in range(score_count, stop_score_count):
      scores[i] = current_scores[i - start_point]
      score_count += 1

  return scores

class WMOF:

# Windowing mass-ratio-variance based outlier factor (WMOF)
# This algorithm is an extension of the mass-ratio-variance outlier factor algorithm (MOF).
# WMOF operates on overlapping windows of fixed size, specified by the user.
# The use of overlapping windows ensures that anomalies occurring at window boundaries are not missed.
# For each window, the MOF score is computed for all data points within the window.

# Parameters-free
# ----------
  def __init__(self):
    self.name='WMOF'
    self.Data = []
    self.decision_scores_ = np.array([], dtype = np.int32)
    self.Anomaly = np.array([])

  def fit(self,Data,Window=100, Overlap_ratio = 0.2):

    '''
    Parameters
    ----------
    Data : numpy array of shape (n_samples, n_features)
        The input samples.
    Window : integer (int)
        number of points for each calculation
        default window size is 100.
    Overlap_ratio : float
        0.0 <= Overlap_ratio <= 0.5
        A Overlap_ratio between window frame.
        default ratio is 0.2
    '''
    '''
    Returns
    -------
    self : object
    '''

  #  Fitted estimator.
    self.Data = Data

  # Calculate mass ratio variance (MOF)
    self.decision_scores_= _Window_Massratio(Data,Window,Overlap_ratio)

  def detectAnomaly(self, theshold):
    '''
    Parameters
    ----------
    theshold : float
        A theshold value for detect anomaly points
    '''
    '''
    Returns
    -------
    idx : numpy array of shape (n_samples,)
        An index array of anomaly points in data
    '''
    # Check data avaliablity
    assert(len(self.Data) != 0)

    if self.decision_scores_.shape[0] == 0:
      self.fit(self.Data)
    idx = np.squeeze(np.argwhere(self.decision_scores_ > theshold))
    self.Anomaly = np.append(self.Anomaly,idx).astype(np.int32)
    self.Anomaly = np.unique(self.Anomaly)

    return idx
  
  def visualize(self):
    '''
    Parameters free
    Visualize data points
    '''
    '''
    Parameters
    ----------
    '''
    '''
    Returns
    -------
    '''
    if self.Data.shape[1] == 3:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot(111, projection='3d')
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], self.Data[:, 2], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    elif self.Data.shape[1] == 2:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot()
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    else :
      print("Cannot visualize dimension space more than 3")
    return self.decision_scores_


import numpy as np
import numba as nb
from numba import jit, objmode, types
from scipy.stats import rankdata
from scipy.spatial.distance import cdist
from sklearn.preprocessing import MinMaxScaler
import matplotlib.pyplot as plt
from .MOF import _point_in_radius

## Scoring function for MAOF
#AAD
@jit(nopython=True)
def _CalAAD(arr, l):
  return np.mean(np.absolute(arr - np.mean(arr)))

#IQR
@jit(nopython=True)
def _CalIQR(arr, l):
  return np.percentile(arr, 75) - np.percentile(arr, 25)

#Weight
@jit(nopython=True)
def _CalWeight(arr, l):# l is lambda for weight
  return l*(np.mean(np.absolute(arr - np.mean(arr)))) + (1-l)*(np.percentile(arr, 75) - np.percentile(arr, 25))

#Range
@jit(nopython=True)
def _CalRange(arr, l):
  return np.ptp(arr)

# Calculate mass ratio matrix
@jit(nopython=True)
def _Massratio(Data,window, funcName, weightLambda):
  # Beware of large numbers, it might overflow python int
    n = len(Data)
    scores = np.zeros(n)

    assert(window > 0)

    # slicing window through data
    for start_point in range(0,n,window):
      stop_point = min(start_point+ window, n)
      Current_Data = Data[start_point : stop_point]
      win_size = stop_point - start_point

      with objmode(window_dm = "i8[:, :]", sort_idx = "i8[:, :]"):
        window_dm = cdist(Current_Data, Data)
        sort_idx = np.argsort(window_dm)

      # when pairwise distance is same, the index is also same
      current_idx = _point_in_radius(window_dm, sort_idx, win_size, n)

      count = np.zeros(win_size, dtype=np.int32)
      mass = np.zeros((win_size, n-1))

      # calculate all current points
      for i in range(n):
        with objmode(dm = "i8[:, :]", remain_idx = "i8[:, :]"):
          dm = cdist([Data[i]], Data)
          remain_idx = np.argsort(dm)

        # when pairwise distance is same, the index is also same
        idx = _point_in_radius(dm, remain_idx, 1, n)

        for j in range(start_point, stop_point):
          if i == j:
            continue
          m = (idx[0][j] + 1.0) / (current_idx[j%window][i] + 1.0 )
          mass[j%window,count[j%window]] = m
          count[j%window] += 1

      # calculate scores
      for i in range(start_point, stop_point):
        arr = mass[i%window]
        scores[i] = funcName(arr, weightLambda)

    return scores

class MAOF:
  '''
  Mass-Ratio-Average-Absolute-Deviation Based Outlier Factor for Anomaly Scoring (MAOF)
  This research extends the mass-ratio-variance outlier factor algorithm (MOF) by exploring other alternative statistical
  dispersion beyond the traditional variance such as range, interquartile range, average absolute deviation, and combination of previous two dispersions.
  '''
  def __init__(self):
    self.name='MAOF'
    self.Data = []
  def fit(self,Data,Window=10000, Function_name = "AAD", Weight_Lambda = 0.5):
    '''
    Parameters
    ----------
    Data : numpy array of shape (n_samples, n_features)
        The input samples.
    Window : integer (int)
        number of points for each calculation
        default window size is 10000.
    Function_name : string (str)
        A type of statistical dispersion that use for scoring.
        Function_name can be 'AAD','IQR', 'Range','Weight'.
        default function is 'AAD'
    Weight_Lambda : float
        0.0 <= Weight_Lambda <= 1.0
        A Value of lambda that use in weight-scoring function.
        score = λ AAD + (1- λ) IQR
        default weight is 0.5
    '''
    '''
    Returns
    -------
    self : object
    '''
    self.Data = Data

    assert(Function_name in ['AAD','IQR','Range','Weight'])
    funcName = _CalAAD
    if Function_name == 'IQR':
      funcName = _CalIQR
    elif Function_name == 'Range':
      funcName = _CalRange
    elif Function_name == 'Weight':
      funcName = _CalWeight
      assert(0.0 <= Weight_Lambda <= 1.0)

    # Calculate Mass-Ratio-Average-Absolute-Deviation (MAOF)
    self.decision_scores_= _Massratio(Data,Window,funcName,Weight_Lambda)

  def visualize(self):
    '''
    Parameters free
    Visualize data points
    '''
    '''
    Parameters
    ----------
    '''
    '''
    Returns
    -------
    '''
    if self.Data.shape[1] == 3:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot(111, projection='3d')
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], self.Data[:, 2], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    elif self.Data.shape[1] == 2:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot()
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    else :
      print("Cannot visualize dimension space more than 3")
    return self.decision_scores_


import numpy as np
import numba as nb
from numba import jit, objmode, types
from scipy.stats import rankdata
from scipy.spatial.distance import cdist
from sklearn.preprocessing import MinMaxScaler
import matplotlib.pyplot as plt

# Calculate number of points in radius
@jit(nopython=True)
def _point_in_radius(dm, sort_idx, win_size, n):
  idx = np.ones((win_size,n), dtype = np.int64)
  for i in range(win_size):
    pre_distance = -1
    point_count = n
    for j in range(n-1, -1, -1):
      obv_point = sort_idx[i][j]
      if dm[i][obv_point] != pre_distance:
        point_count = j
        pre_distance = dm[i][obv_point]
      idx[i][obv_point] = point_count
  return idx

# Calculate mass ratio matrix
@jit(nopython=True)
def _Var_Massratio(Data,window):
  # Beware of large numbers, it might overflow python int
    n = len(Data)
    mass = np.zeros(n)
    mass2 = np.zeros(n)
    assert(window > 0)

    # slicing window through data
    for start_point in range(0,n,window):
      stop_point = min(start_point+ window, n)
      Current_Data = Data[start_point : stop_point]
      win_size = stop_point - start_point

      with objmode(window_dm = "i8[:, :]", sort_idx = "i8[:, :]"):
        window_dm = cdist(Current_Data, Data)
        sort_idx = np.argsort(window_dm)

      # when pairwise distance is same, the index is also same
      current_idx = _point_in_radius(window_dm, sort_idx, win_size, n)

      # calculate all current points
      for i in range(start_point, stop_point):
        for j in range(i+1, stop_point):
          m = (current_idx[j%window][i]*1.0 + 1)/ (current_idx[i%window][j] + 1)
          mass[i] += m
          mass2[i] += m**2
          mass[j] += 1/m
          mass2[j] += 1/m**2

      # calculate remaining points
      for i in range(stop_point,n):
        with objmode(dm = "i8[:, :]", sort_remain = "i8[:, :]"):
          dm = cdist([Data[i]], Data)
          sort_remain = np.argsort(dm)

        # when pairwise distance is same, the index is also same
        idx = _point_in_radius(dm, sort_remain, 1, n)

        for j in range(start_point, stop_point):
          m = (current_idx[j%window][i]*1.0 + 1 )/ (idx[0][j] + 1)
          mass[i] += m
          mass2[i] += m**2
          mass[j] += 1/m
          mass2[j] += 1/m**2

    var = mass2/(n-1)-(mass/(n-1))**2
    return var

class MOF:
  '''
  Mass ratio variance-based outlier factor (MOF)
  the outlier score of each data point is called MOF.
  It measures the global deviation of density given sample with respect to other data points.
  it is global in the outlier score depend on how isolated
  data point is with respect to all data points in the data set.
  the variance of mass ratio can identify data points that have a substantially
  lower density compared to other data points.
  These are considered outliers.
  '''
  # Parameters-free
  # ----------
  def __init__(self):
    self.name='MOF'
    self.Data = []

  def fit(self,Data, Window = 10000):

    '''
    Parameters
    ----------
    Data : numpy array of shape (n_samples, n_features)
        The input samples.
    window : integer (int)
        window size for calculation.
        default window size is 10000.
    '''
    '''
    Returns
    -------
    self : object
    '''
#  Fitted estimator.
    self.Data =Data

# Calculate mass ratio variance (MOF)
    self.decision_scores_= _Var_Massratio(Data,Window)

# ----------------
  def visualize(self):
    '''
    Parameters free
    Visualize data points
    '''
    '''
    Parameters
    ----------
    '''
    '''
    Returns
    -------
    '''
    if self.Data.shape[1] == 3:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot(111, projection='3d')
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], self.Data[:, 2], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    elif self.Data.shape[1] == 2:
      fig = plt.figure(figsize=(15, 15))
      ax = fig.add_subplot()
      p = ax.scatter(self.Data[:, 0], self.Data[:, 1], c = np.log(self.decision_scores_+0.00001), cmap='jet')
      fig.colorbar(p)
    else :
      print("Cannot visualize dimension space more than 3")
    return self.decision_scores_
